# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    bint       PyBytes_CheckExact(object)
    char*      PyBytes_AsString(object) except NULL
    Py_ssize_t PyBytes_Size(object) except -1
    object     PyBytes_FromStringAndSize(char*, Py_ssize_t)

# -----------------------------------------------------------------------------

cdef object PyPickle_dumps = None
cdef object PyPickle_loads = None
cdef object PyPickle_PROTOCOL = None
cdef object PyPickle_THRESHOLD = 1024**2 // 4  # 0.25 MiB

from pickle import dumps as PyPickle_dumps
from pickle import loads as PyPickle_loads
from pickle import HIGHEST_PROTOCOL as PyPickle_PROTOCOL

if Py_GETENV(b"MPI4PY_PICKLE_PROTOCOL") != NULL:
    PyPickle_PROTOCOL = int(Py_GETENV(b"MPI4PY_PICKLE_PROTOCOL"))

if Py_GETENV(b"MPI4PY_PICKLE_THRESHOLD") != NULL:
    PyPickle_THRESHOLD = int(Py_GETENV(b"MPI4PY_PICKLE_THRESHOLD"))

cdef class Pickle:
    """
    Pickle/unpickle Python objects.
    """

    cdef object ob_dumps
    cdef object ob_loads
    cdef object ob_PROTO
    cdef object ob_THRES

    def __cinit__(self, *args, **kwargs):
        <void> args    # unused
        <void> kwargs  # unused
        self.ob_dumps = PyPickle_dumps
        self.ob_loads = PyPickle_loads
        self.ob_PROTO = PyPickle_PROTOCOL
        self.ob_THRES = PyPickle_THRESHOLD

    def __init__(
        self,
        dumps: Callable[[Any, int], bytes] | None = None,
        loads: Callable[[Buffer], Any] | None = None,
        protocol: int | None = None,
        threshold: int | None = None,
    ) -> None:
        if dumps is None:
            dumps = PyPickle_dumps
        if loads is None:
            loads = PyPickle_loads
        if protocol is None:
            if dumps is PyPickle_dumps:
                protocol = PyPickle_PROTOCOL
        if threshold is None:
            threshold = PyPickle_THRESHOLD
        self.ob_dumps = dumps
        self.ob_loads = loads
        self.ob_PROTO = protocol
        self.ob_THRES = threshold

    def dumps(
        self,
        obj: Any,
    ) -> bytes:
        """
        Serialize object to pickle data stream.
        """
        return cdumps(self, obj)

    def loads(
        self,
        data: Buffer,
    ) -> Any:
        """
        Deserialize object from pickle data stream.
        """
        return cloads(self, data)

    def dumps_oob(
        self,
        obj: Any,
    ) -> tuple[bytes, list[buffer]]:
        """
        Serialize object to pickle data stream and out-of-band buffers.
        """
        return cdumps_oob(self, obj)

    def loads_oob(
        self,
        data: Buffer,
        buffers: Iterable[Buffer],
    ) -> Any:
        """
        Deserialize object from pickle data stream and out-of-band buffers.
        """
        return cloads_oob(self, data, buffers)

    property PROTOCOL:
        """Protocol version."""
        def __get__(self) -> int | None:
            return self.ob_PROTO

        def __set__(self, protocol: int | None):
            if protocol is None:
                if self.ob_dumps is PyPickle_dumps:
                    protocol = PyPickle_PROTOCOL
            self.ob_PROTO = protocol

    property THRESHOLD:
        """Out-of-band threshold."""
        def __get__(self) -> int:
            return self.ob_THRES

        def __set__(self, threshold: int | None):
            if threshold is None:
                threshold = PyPickle_THRESHOLD
            self.ob_THRES = threshold


cdef Pickle PyMPI_PICKLE = Pickle()
pickle = PyMPI_PICKLE

# -----------------------------------------------------------------------------

cdef object get_buffer_callback(list buffers, Py_ssize_t threshold):
    def buffer_callback(ob):
        cdef buffer buf = getbuffer(ob, 1, 0)
        if buf.view.len >= threshold:
            buffers.append(buf)
            return False
        else:
            return True
    return buffer_callback

cdef object cdumps_oob(Pickle pkl, object obj):
    cdef object pkl_dumps = pkl.ob_dumps
    cdef object protocol = pkl.ob_PROTO
    if protocol is None:
        protocol = PyPickle_PROTOCOL         # ~> uncovered
    protocol = max(protocol, 5)
    cdef list buffers = []
    cdef Py_ssize_t threshold = pkl.ob_THRES
    cdef object buf_cb = get_buffer_callback(buffers, threshold)
    cdef object data = pkl_dumps(obj, protocol, buffer_callback=buf_cb)
    return data, buffers

cdef object cloads_oob(Pickle pkl, object data, object buffers):
    cdef object pkl_loads = pkl.ob_loads
    return pkl_loads(data, buffers=buffers)

# -----------------------------------------------------------------------------

cdef object cdumps(Pickle pkl, object obj):
    if pkl.ob_PROTO is not None:
        return pkl.ob_dumps(obj, pkl.ob_PROTO)
    else:
        return pkl.ob_dumps(obj)

cdef object cloads(Pickle pkl, object buf):
    return pkl.ob_loads(buf)


cdef object pickle_dump(Pickle pkl, object obj, void **p, MPI_Count *n):
    cdef object buf = cdumps(pkl, obj)
    p[0] = PyBytes_AsString(buf)
    n[0] = PyBytes_Size(buf)
    return buf

cdef object pickle_load(Pickle pkl, void *p, MPI_Count n):
    if p == NULL or n == 0: return None
    return cloads(pkl, mpibuf(p, n))


cdef object pickle_dumpv(
    Pickle pkl, object obj,
    void **p, int n, MPI_Count cnt[], MPI_Aint dsp[],
):
    cdef Py_ssize_t m = n
    cdef object items
    if obj is None: items = [None] * m
    else:           items = list(obj)
    m = len(items)
    if m != n:
        raise ValueError(f"expecting {n} items, got {m}")
    cdef MPI_Count c = 0
    cdef MPI_Aint  d = 0
    for i in range(m):
        items[i] = pickle_dump(pkl, items[i], p, &c)
        cnt[i] = c
        dsp[i] = d
        d = d + <MPI_Aint> c
    cdef object buf = b''.join(items)
    p[0] = PyBytes_AsString(buf)
    return buf

cdef object pickle_loadv(
    Pickle pkl,
    void *p, int n, MPI_Count cnt[], MPI_Aint dsp[],
):
    cdef Py_ssize_t m = n
    cdef object items = [None] * m
    if p == NULL: return items
    for i in range(m):
        items[i] = pickle_load(pkl, <char*>p + dsp[i], cnt[i])
    return items


cdef object pickle_alloc(void **p, MPI_Count n):
    cdef object buf = PyBytes_FromStringAndSize(NULL, <MPI_Aint>n)
    p[0] = PyBytes_AsString(buf)
    return buf

cdef object pickle_allocv(void **p, int n, MPI_Count cnt[], MPI_Aint dsp[]):
    cdef MPI_Count d=0
    for i in range(n):
        dsp[i] = <MPI_Aint> d
        d += cnt[i]
    return pickle_alloc(p, d)


cdef object allocate_count_displ(int n, MPI_Count **p, MPI_Aint **q):
    cdef object mem1 = allocate(n, sizeof(MPI_Count), p)
    cdef object mem2 = allocate(n, sizeof(MPI_Aint),  q)
    return (mem1, mem2)

# -----------------------------------------------------------------------------

cdef object PyMPI_send(object obj, int dest, int tag,
                       MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    #
    cdef object unuseds = None
    if dest != MPI_PROC_NULL:
        unuseds = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Send_c(
        sbuf, scount, stype,
        dest, tag, comm) )
    return None


cdef object PyMPI_bsend(object obj, int dest, int tag,
                        MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    #
    cdef object unuseds = None
    if dest != MPI_PROC_NULL:
        unuseds = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Bsend_c(
        sbuf, scount, stype,
        dest, tag, comm) )
    return None


cdef object PyMPI_ssend(object obj, int dest, int tag,
                        MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    #
    cdef object unuseds = None
    if dest != MPI_PROC_NULL:
        unuseds = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Ssend_c(
        sbuf, scount, stype, dest, tag, comm) )
    return None

# -----------------------------------------------------------------------------

cdef object PyMPI_recv_obarg(object obj, int source, int tag,
                             MPI_Comm comm, MPI_Status *status):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *rbuf = NULL
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    cdef MPI_Status rsts
    cdef object   rmsg = None
    cdef MPI_Aint rlen = 0
    #
    PyErr_WarnFormat(
        UserWarning, 1, b"%s",
        b"the 'buf' argument is deprecated",
    )
    #
    if source != MPI_PROC_NULL:
        if is_integral(obj):
            rcount = <MPI_Count> PyNumber_Index(obj)
            rmsg = pickle_alloc(&rbuf, rcount)
        else:
            rmsg = asbuffer_w(obj, &rbuf, &rlen)
            rcount = <MPI_Count> rlen
        if status == MPI_STATUS_IGNORE:
            status = &rsts
        <void> rmsg
    with nogil:
        CHKERR( MPI_Recv_c(
            rbuf, rcount, rtype,
            source, tag, comm, status) )
        if source != MPI_PROC_NULL:
            CHKERR( MPI_Get_count_c(status, rtype, &rcount) )
    #
    if rcount <= 0: return None
    return pickle_load(pickle, rbuf, rcount)


cdef object PyMPI_recv_match(object obj, int source, int tag,
                             MPI_Comm comm, MPI_Status *status):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *rbuf = NULL
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef MPI_Message match = MPI_MESSAGE_NULL
    cdef MPI_Status rsts
    <void> obj  # unused
    #
    with nogil:
        CHKERR( MPI_Mprobe(source, tag, comm, &match, &rsts) )
        CHKERR( MPI_Get_count_c(&rsts, rtype, &rcount) )
    cdef object unusedr = pickle_alloc(&rbuf, rcount)
    with nogil:
        CHKERR( MPI_Mrecv_c(
            rbuf, rcount, rtype, &match, status) )
    #
    if rcount <= 0: return None
    return pickle_load(pickle, rbuf, rcount)


cdef object PyMPI_recv_probe(object obj, int source, int tag,
                             MPI_Comm comm, MPI_Status *status):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *rbuf = NULL
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef MPI_Status rsts
    cdef object unusedr
    <void> obj  # unused
    #
    with PyMPI_Lock(comm, "recv"):
        with nogil:
            CHKERR( MPI_Probe(source, tag, comm, &rsts) )
            CHKERR( MPI_Get_count_c(&rsts, rtype, &rcount) )
            CHKERR( MPI_Status_get_source(&rsts, &source) )
            CHKERR( MPI_Status_get_tag(&rsts, &tag) )
        unusedr = pickle_alloc(&rbuf, rcount)
        with nogil:
            CHKERR( MPI_Recv_c(
                rbuf, rcount, rtype,
                source, tag, comm, status) )
    #
    if rcount <= 0: return None
    return pickle_load(pickle, rbuf, rcount)


cdef object PyMPI_recv(object obj, int source, int tag,
                       MPI_Comm comm, MPI_Status *status):
    if obj is not None:
        return PyMPI_recv_obarg(obj, source, tag, comm, status)
    elif options.recv_mprobe:
        return PyMPI_recv_match(obj, source, tag, comm, status)
    else:
        return PyMPI_recv_probe(obj, source, tag, comm, status)

# -----------------------------------------------------------------------------

cdef object PyMPI_isend(object obj, int dest, int tag,
                        MPI_Comm comm, MPI_Request *request):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    #
    cdef object smsg = None
    if dest != MPI_PROC_NULL:
        smsg = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Isend_c(
        sbuf, scount, stype,
        dest, tag, comm, request) )
    return smsg


cdef object PyMPI_ibsend(object obj, int dest, int tag,
                         MPI_Comm comm, MPI_Request *request):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    #
    cdef object smsg = None
    if dest != MPI_PROC_NULL:
        smsg = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Ibsend_c(
        sbuf, scount, stype,
        dest, tag, comm, request) )
    return smsg


cdef object PyMPI_issend(object obj, int dest, int tag,
                         MPI_Comm comm, MPI_Request *request):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    #
    cdef object smsg = None
    if dest != MPI_PROC_NULL:
        smsg = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Issend_c(
        sbuf, scount, stype,
        dest, tag, comm, request) )
    return smsg


cdef object PyMPI_irecv(object obj, int source, int tag,
                        MPI_Comm comm, MPI_Request *request):
    #
    cdef void *rbuf = NULL
    cdef MPI_Aint rlen = 0
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef object rmsg = None
    if source != MPI_PROC_NULL:
        if obj is None:
            rcount = options.irecv_bufsz
            obj = pickle_alloc(&rbuf, rcount)
            rmsg = asbuffer_r(obj, NULL, NULL)
        elif is_integral(obj):
            rcount = <MPI_Count> PyNumber_Index(obj)
            obj = pickle_alloc(&rbuf, rcount)
            rmsg = asbuffer_r(obj, NULL, NULL)
        else:
            rmsg = asbuffer_w(obj, &rbuf, &rlen)
            rcount = <MPI_Count> rlen
        rmsg = PyMPI_wrap_buffer(<buffer>rmsg)
    with nogil: CHKERR( MPI_Irecv_c(
        rbuf, rcount, rtype,
        source, tag, comm, request) )
    return rmsg

# -----------------------------------------------------------------------------

cdef object PyMPI_sendrecv(object sobj, int dest,   int sendtag,
                           object robj, int source, int recvtag,
                           MPI_Comm comm, MPI_Status *status):
    cdef MPI_Request request = MPI_REQUEST_NULL
    sobj = PyMPI_isend(sobj, dest,   sendtag, comm, &request)
    robj = PyMPI_recv (robj, source, recvtag, comm, status)
    with nogil: CHKERR( MPI_Wait(&request, MPI_STATUS_IGNORE) )
    return robj

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_req_buf:
    cdef buffer buf

cdef inline object PyMPI_wrap_buffer(buffer buf):
    cdef _p_req_buf ob = _p_req_buf.__new__(_p_req_buf)
    ob.buf = buf
    return ob

cdef inline object PyMPI_load_buffer(_p_req_buf ob, MPI_Status *status):
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    CHKERR( MPI_Get_count_c(status, rtype, &rcount) )
    if rcount <= 0: return None
    cdef Pickle pickle = PyMPI_PICKLE
    cdef void *rbuf = ob.buf.view.buf
    return pickle_load(pickle, rbuf, rcount)


@cython.final
@cython.internal
cdef class _p_req_obj:
    cdef object obj

cdef inline object PyMPI_wrap_object(object obj):
    cdef _p_req_obj ob = _p_req_obj.__new__(_p_req_obj)
    ob.obj = obj
    return ob

cdef inline object PyMPI_load_object(_p_req_obj ob):
    return ob.obj


cdef inline object PyMPI_load(object ob, MPI_Status *status):
    if type(ob) is _p_req_buf:
        return PyMPI_load_buffer(<_p_req_buf>ob, status)
    if type(ob) is _p_req_obj:
        return PyMPI_load_object(<_p_req_obj>ob)
    return None

# -----------------------------------------------------------------------------

cdef object PyMPI_wait(Request request, Status status):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.set_request(request)
    rs.set_status(status)
    with nogil: CHKERR( MPI_Wait(&request.ob_mpi, rs.status) )
    return rs.get_result()

cdef object PyMPI_test(Request request, int *flag, Status status):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.set_request(request)
    rs.set_status(status)
    with nogil: CHKERR( MPI_Test(&request.ob_mpi, flag, rs.status) )
    if not flag[0]:
        return None
    return rs.get_result()

cdef object PyMPI_waitany(requests, int *index, Status status):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.acquire(requests)
    rs.set_status(status)
    cdef object obj = None
    try:
        with nogil: CHKERR( MPI_Waitany(
            rs.count, rs.requests, index, rs.status) )
        if index[0] != MPI_UNDEFINED:
            obj = rs.get_object(index[0])
    finally:
        rs.release()
    return obj

cdef object PyMPI_testany(requests, int *index, int *flag, Status status):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.acquire(requests)
    rs.set_status(status)
    cdef object obj = None
    try:
        with nogil: CHKERR( MPI_Testany(
            rs.count, rs.requests, index, flag, rs.status) )
        if index[0] != MPI_UNDEFINED and flag[0]:
            obj = rs.get_object(index[0])
    finally:
        rs.release()
    return obj

cdef object PyMPI_waitall(requests, statuses):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.acquire(requests)
    rs.add_statuses()
    cdef object objects = None
    try:
        with nogil: CHKERR( MPI_Waitall(
            rs.count, rs.requests, rs.statuses) )
        objects = rs.get_objects()
    finally:
        rs.release(statuses)
    return objects

cdef object PyMPI_testall(requests, int *flag, statuses):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.acquire(requests)
    rs.add_statuses()
    cdef object objects = None
    try:
        with nogil: CHKERR( MPI_Testall(
            rs.count, rs.requests, flag, rs.statuses) )
        if flag[0]:
            objects = rs.get_objects()
    finally:
        rs.release(statuses)
    return objects

cdef object PyMPI_waitsome(requests, statuses):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.acquire(requests)
    rs.add_indices()
    rs.add_statuses()
    cdef object indices = None
    cdef object objects = None
    try:
        with nogil: CHKERR( MPI_Waitsome(
            rs.count, rs.requests, &rs.outcount, rs.indices, rs.statuses) )
        indices = rs.get_indices()
        objects = rs.get_objects()
    finally:
        rs.release(statuses)
    return (indices, objects)

cdef object PyMPI_testsome(requests, statuses):
    cdef _p_rs rs = _p_rs.__new__(_p_rs)
    rs.acquire(requests)
    rs.add_indices()
    rs.add_statuses()
    cdef object indices = None
    cdef object objects = None
    try:
        with nogil: CHKERR( MPI_Testsome(
            rs.count, rs.requests, &rs.outcount, rs.indices, rs.statuses) )
        indices = rs.get_indices()
        objects = rs.get_objects()
    finally:
        rs.release(statuses)
    return (indices, objects)

# -----------------------------------------------------------------------------

cdef object PyMPI_probe(int source, int tag,
                        MPI_Comm comm, MPI_Status *status):
    with nogil: CHKERR( MPI_Probe(source, tag, comm, status) )
    return True

cdef object PyMPI_iprobe(int source, int tag,
                         MPI_Comm comm, MPI_Status *status):
    cdef int flag = 0
    with nogil: CHKERR( MPI_Iprobe(source, tag, comm, &flag, status) )
    return <bint>flag

cdef object PyMPI_mprobe(int source, int tag, MPI_Comm comm,
                         MPI_Message *message, MPI_Status *status):
    cdef void* rbuf = NULL
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    cdef MPI_Status rsts
    if (status == MPI_STATUS_IGNORE): status = &rsts
    with nogil: CHKERR( MPI_Mprobe(source, tag, comm, message, status) )
    if message[0] == MPI_MESSAGE_NO_PROC: return None
    CHKERR( MPI_Get_count_c(status, rtype, &rcount) )
    cdef object rmsg = pickle_alloc(&rbuf, rcount)
    return rmsg

cdef object PyMPI_improbe(int source, int tag, MPI_Comm comm, int *flag,
                          MPI_Message *message, MPI_Status *status):
    cdef void* rbuf = NULL
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    cdef MPI_Status rsts
    if (status == MPI_STATUS_IGNORE): status = &rsts
    with nogil: CHKERR( MPI_Improbe(source, tag, comm, flag, message, status) )
    if flag[0] == 0 or message[0] == MPI_MESSAGE_NO_PROC: return None
    CHKERR( MPI_Get_count_c(status, rtype, &rcount) )
    cdef object rmsg = pickle_alloc(&rbuf, rcount)
    return rmsg

cdef object PyMPI_mrecv(object rmsg,
                        MPI_Message *message, MPI_Status *status):
    cdef Pickle pickle = PyMPI_PICKLE
    cdef void* rbuf = NULL
    cdef MPI_Aint rlen = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    if message[0] == MPI_MESSAGE_NO_PROC:
        rmsg = None
    elif rmsg is None:
        pass
    elif PyBytes_CheckExact(rmsg):
        rmsg = asbuffer_r(rmsg, &rbuf, &rlen)
    else:
        rmsg = asbuffer_w(rmsg, &rbuf, &rlen)  # ~> unreachable
    cdef MPI_Count rcount = <MPI_Count> rlen
    with nogil: CHKERR( MPI_Mrecv_c(
        rbuf, rcount, rtype, message, status) )
    rmsg = pickle_load(pickle, rbuf, rcount)
    return rmsg

cdef object PyMPI_imrecv(object rmsg,
                         MPI_Message *message, MPI_Request *request):
    cdef void* rbuf = NULL
    cdef MPI_Aint rlen = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    if message[0] == MPI_MESSAGE_NO_PROC:
        rmsg = None
    elif rmsg is None:
        pass
    elif PyBytes_CheckExact(rmsg):
        rmsg = asbuffer_r(rmsg, &rbuf, &rlen)
    else:
        rmsg = asbuffer_w(rmsg, &rbuf, &rlen)  # ~> unreachable
    if rmsg is not None:
        rmsg = PyMPI_wrap_buffer(<buffer>rmsg)
    cdef MPI_Count rcount = <MPI_Count> rlen
    with nogil: CHKERR( MPI_Imrecv_c(
        rbuf, rcount, rtype, message, request) )
    return rmsg

# -----------------------------------------------------------------------------

cdef object PyMPI_barrier(MPI_Comm comm):
    with nogil: CHKERR( MPI_Barrier(comm) )
    return None


cdef object PyMPI_bcast(object obj, int root, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *buf = NULL
    cdef MPI_Count count = 0
    cdef MPI_Datatype dtype = MPI_BYTE
    #
    cdef int dosend=0, dorecv=0
    cdef int inter=0, rank=0
    CHKERR( MPI_Comm_test_inter(comm, &inter) )
    if inter:
        if root == MPI_PROC_NULL:
            dosend, dorecv = 0, 0
        elif root == MPI_ROOT:
            dosend, dorecv = 1, 0
        else:
            dosend, dorecv = 0, 1
    else:
        CHKERR( MPI_Comm_rank(comm, &rank) )
        if root == rank:
            dosend, dorecv = 1, 1
        else:
            dosend, dorecv = 0, 1
    #
    cdef object smsg = None
    cdef object rmsg = None
    #
    if dosend: smsg = pickle_dump(pickle, obj, &buf, &count)
    if dosend and dorecv: rmsg = smsg
    with PyMPI_Lock(comm, "bcast"):
        with nogil: CHKERR( MPI_Bcast_c(
            &count, 1, MPI_COUNT,
            root, comm) )
        if dorecv and not dosend:
            rmsg = pickle_alloc(&buf, count)
        with nogil: CHKERR( MPI_Bcast_c(
            buf, count, dtype,
            root, comm) )
    if dorecv: rmsg = pickle_load(pickle, buf, count)
    #
    return rmsg


cdef object PyMPI_gather(object sendobj, int root, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    cdef void *rbuf = NULL
    cdef MPI_Count *rcounts = NULL
    cdef MPI_Aint  *rdispls = NULL
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef int dosend=0, dorecv=0
    cdef int inter=0, size=0, rank=0
    CHKERR( MPI_Comm_test_inter(comm, &inter) )
    if inter:
        CHKERR( MPI_Comm_remote_size(comm, &size) )
        if root == MPI_PROC_NULL:
            dosend, dorecv = 0, 0
        elif root == MPI_ROOT:
            dosend, dorecv = 0, 1
        else:
            dosend, dorecv = 1, 0
    else:
        CHKERR( MPI_Comm_size(comm, &size) )
        CHKERR( MPI_Comm_rank(comm, &rank) )
        if root == rank:
            dosend, dorecv = 1, 1
        else:
            dosend, dorecv = 1, 0
    #
    cdef object unuseds = None
    cdef object rmsg = None
    cdef object unused1
    #
    if dorecv: unused1 = allocate_count_displ(size, &rcounts, &rdispls)
    if dosend: unuseds = pickle_dump(pickle, sendobj, &sbuf, &scount)
    with PyMPI_Lock(comm, "gather"):
        with nogil: CHKERR( MPI_Gather_c(
            &scount, 1, MPI_COUNT,
            rcounts, 1, MPI_COUNT,
            root, comm) )
        if dorecv: rmsg = pickle_allocv(&rbuf, size, rcounts, rdispls)
        with nogil: CHKERR( MPI_Gatherv_c(
            sbuf, scount,           stype,
            rbuf, rcounts, rdispls, rtype,
            root, comm) )
    if dorecv: rmsg = pickle_loadv(pickle, rbuf, size, rcounts, rdispls)
    #
    return rmsg


cdef object PyMPI_scatter(object sendobj, int root, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count *scounts = NULL
    cdef MPI_Aint  *sdispls = NULL
    cdef MPI_Datatype stype = MPI_BYTE
    cdef void *rbuf = NULL
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef int dosend=0, dorecv=0
    cdef int inter=0, size=0, rank=0
    CHKERR( MPI_Comm_test_inter(comm, &inter) )
    if inter:
        CHKERR( MPI_Comm_remote_size(comm, &size) )
        if root == MPI_PROC_NULL:
            dosend, dorecv = 0, 0
        elif root == MPI_ROOT:
            dosend, dorecv = 1, 0
        else:
            dosend, dorecv = 0, 1
    else:
        CHKERR( MPI_Comm_size(comm, &size) )
        CHKERR( MPI_Comm_rank(comm, &rank) )
        if root == rank:
            dosend, dorecv = 1, 1
        else:
            dosend, dorecv = 0, 1
    #
    cdef object unuseds = None
    cdef object rmsg = None
    cdef object unused1
    #
    if dosend: unused1 = allocate_count_displ(size, &scounts, &sdispls)
    if dosend: unuseds = pickle_dumpv(pickle, sendobj, &sbuf, size, scounts, sdispls)
    with PyMPI_Lock(comm, "scatter"):
        with nogil: CHKERR( MPI_Scatter_c(
            scounts, 1, MPI_COUNT,
            &rcount, 1, MPI_COUNT,
            root, comm) )
        if dorecv: rmsg = pickle_alloc(&rbuf, rcount)
        with nogil: CHKERR( MPI_Scatterv_c(
            sbuf, scounts, sdispls, stype,
            rbuf, rcount,           rtype,
            root, comm) )
    if dorecv: rmsg = pickle_load(pickle, rbuf, rcount)
    #
    return rmsg


cdef object PyMPI_allgather(object sendobj, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    cdef void *rbuf = NULL
    cdef MPI_Count *rcounts = NULL
    cdef MPI_Aint  *rdispls = NULL
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef int inter=0, size=0
    CHKERR( MPI_Comm_test_inter(comm, &inter) )
    if inter:
        CHKERR( MPI_Comm_remote_size(comm, &size) )
    else:
        CHKERR( MPI_Comm_size(comm, &size) )
    #
    cdef object unuseds = None
    cdef object rmsg = None
    cdef object unused1
    #
    unused1 = allocate_count_displ(size, &rcounts, &rdispls)
    unuseds = pickle_dump(pickle, sendobj, &sbuf, &scount)
    with PyMPI_Lock(comm, "allgather"):
        with nogil: CHKERR( MPI_Allgather_c(
            &scount, 1, MPI_COUNT,
            rcounts, 1, MPI_COUNT,
            comm) )
        rmsg = pickle_allocv(&rbuf, size, rcounts, rdispls)
        with nogil: CHKERR( MPI_Allgatherv_c(
            sbuf, scount,           stype,
            rbuf, rcounts, rdispls, rtype,
            comm) )
    rmsg = pickle_loadv(pickle, rbuf, size, rcounts, rdispls)
    #
    return rmsg


cdef object PyMPI_alltoall(object sendobj, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count *scounts = NULL
    cdef MPI_Aint  *sdispls = NULL
    cdef MPI_Datatype stype = MPI_BYTE
    cdef void *rbuf = NULL
    cdef MPI_Count *rcounts = NULL
    cdef MPI_Aint  *rdispls = NULL
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef int inter=0, size=0
    CHKERR( MPI_Comm_test_inter(comm, &inter) )
    if inter:
        CHKERR( MPI_Comm_remote_size(comm, &size) )
    else:
        CHKERR( MPI_Comm_size(comm, &size) )
    #
    cdef object unuseds = None
    cdef object rmsg = None
    cdef object unused1, unused2
    #
    unused1 = allocate_count_displ(size, &scounts, &sdispls)
    unused2 = allocate_count_displ(size, &rcounts, &rdispls)
    unuseds = pickle_dumpv(pickle, sendobj, &sbuf, size, scounts, sdispls)
    with PyMPI_Lock(comm, "alltoall"):
        with nogil: CHKERR( MPI_Alltoall_c(
            scounts, 1, MPI_COUNT,
            rcounts, 1, MPI_COUNT,
            comm) )
        rmsg = pickle_allocv(&rbuf, size, rcounts, rdispls)
        with nogil: CHKERR( MPI_Alltoallv_c(
            sbuf, scounts, sdispls, stype,
            rbuf, rcounts, rdispls, rtype,
            comm) )
    rmsg = pickle_loadv(pickle, rbuf, size, rcounts, rdispls)
    #
    return rmsg


cdef object PyMPI_neighbor_allgather(object sendobj, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    cdef void *rbuf = NULL
    cdef MPI_Count *rcounts = NULL
    cdef MPI_Aint  *rdispls = NULL
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef int rsize=0
    comm_neighbors_count(comm, &rsize, NULL)
    #
    cdef object unuseds = None
    cdef object rmsg = None
    cdef object unused1
    #
    unused1 = allocate_count_displ(rsize, &rcounts, &rdispls)
    for i in range(rsize): rcounts[i] = 0
    unuseds = pickle_dump(pickle, sendobj, &sbuf, &scount)
    with PyMPI_Lock(comm, "neighbor_allgather"):
        with nogil: CHKERR( MPI_Neighbor_allgather_c(
            &scount, 1, MPI_COUNT,
            rcounts, 1, MPI_COUNT,
            comm) )
        rmsg = pickle_allocv(&rbuf, rsize, rcounts, rdispls)
        with nogil: CHKERR( MPI_Neighbor_allgatherv_c(
            sbuf, scount,           stype,
            rbuf, rcounts, rdispls, rtype,
            comm) )
    rmsg = pickle_loadv(pickle, rbuf, rsize, rcounts, rdispls)
    #
    return rmsg


cdef object PyMPI_neighbor_alltoall(object sendobj, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    #
    cdef void *sbuf = NULL
    cdef MPI_Count *scounts = NULL
    cdef MPI_Aint  *sdispls = NULL
    cdef MPI_Datatype stype = MPI_BYTE
    cdef void *rbuf = NULL
    cdef MPI_Count *rcounts = NULL
    cdef MPI_Aint  *rdispls = NULL
    cdef MPI_Datatype rtype = MPI_BYTE
    #
    cdef int ssize=0, rsize=0
    comm_neighbors_count(comm, &rsize, &ssize)
    #
    cdef object unuseds = None
    cdef object rmsg = None
    cdef object unused1, unused2
    #
    unused1 = allocate_count_displ(ssize, &scounts, &sdispls)
    unused2 = allocate_count_displ(rsize, &rcounts, &rdispls)
    for i in range(rsize): rcounts[i] = 0
    unuseds = pickle_dumpv(pickle, sendobj, &sbuf, ssize, scounts, sdispls)
    with PyMPI_Lock(comm, "neighbor_alltoall"):
        with nogil: CHKERR( MPI_Neighbor_alltoall_c(
            scounts, 1, MPI_COUNT,
            rcounts, 1, MPI_COUNT,
            comm) )
        rmsg = pickle_allocv(&rbuf, rsize, rcounts, rdispls)
        with nogil: CHKERR( MPI_Neighbor_alltoallv_c(
            sbuf, scounts, sdispls, stype,
            rbuf, rcounts, rdispls, rtype,
            comm) )
    rmsg = pickle_loadv(pickle, rbuf, rsize, rcounts, rdispls)
    #
    return rmsg

# -----------------------------------------------------------------------------

cdef inline object _py_reduce(object seq, object op):
    if seq is None: return None
    cdef Py_ssize_t i, n = len(seq)
    cdef object res = seq[0]
    for i in range(1, n):
        res = op(res, seq[i])
    return res

cdef inline object _py_scan(object seq, object op):
    if seq is None: return None
    cdef Py_ssize_t i, n = len(seq)
    for i in range(1, n):
        seq[i] = op(seq[i-1], seq[i])
    return seq

cdef inline object _py_exscan(object seq, object op):
    if seq is None: return None
    seq = _py_scan(seq, op)
    seq.pop(-1)
    seq.insert(0, None)
    return seq

cdef object PyMPI_reduce_naive(object sendobj, object op,
                               int root, MPI_Comm comm):
    cdef object items = PyMPI_gather(sendobj, root, comm)
    return _py_reduce(items, op)

cdef object PyMPI_allreduce_naive(object sendobj, object op, MPI_Comm comm):
    cdef object items = PyMPI_allgather(sendobj, comm)
    return _py_reduce(items, op)

cdef object PyMPI_scan_naive(object sendobj, object op, MPI_Comm comm):
    cdef object items = PyMPI_gather(sendobj, 0, comm)
    items = _py_scan(items, op)
    return PyMPI_scatter(items, 0, comm)

cdef object PyMPI_exscan_naive(object sendobj, object op, MPI_Comm comm):
    cdef object items = PyMPI_gather(sendobj, 0, comm)
    items = _py_exscan(items, op)
    return PyMPI_scatter(items, 0, comm)

# ---

cdef object PyMPI_copy(object obj):
    cdef Pickle pickle = PyMPI_PICKLE
    cdef void *buf = NULL
    cdef MPI_Count count = 0
    obj = pickle_dump(pickle, obj, &buf, &count)
    return pickle_load(pickle, buf, count)

cdef object PyMPI_send_p2p(object obj, int dst, int tag, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    cdef void *sbuf = NULL
    cdef MPI_Count scount = 0
    cdef MPI_Datatype stype = MPI_BYTE
    cdef object unuseds = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Send_c(&scount, 1, MPI_COUNT, dst, tag, comm) )
    with nogil: CHKERR( MPI_Send_c(sbuf, scount, stype, dst, tag, comm) )
    return None

cdef object PyMPI_recv_p2p(int src, int tag, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    cdef void *rbuf = NULL
    cdef MPI_Count rcount = 0
    cdef MPI_Datatype rtype = MPI_BYTE
    cdef MPI_Status *status = MPI_STATUS_IGNORE
    with nogil: CHKERR( MPI_Recv_c(&rcount, 1, MPI_COUNT, src, tag, comm, status) )
    cdef object unusedr = pickle_alloc(&rbuf, rcount)
    with nogil: CHKERR( MPI_Recv_c(rbuf, rcount, rtype, src, tag, comm, status) )
    return pickle_load(pickle, rbuf, rcount)

cdef object PyMPI_sendrecv_p2p(object obj,
                               int dst, int stag,
                               int src, int rtag,
                               MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    cdef void *sbuf = NULL, *rbuf = NULL
    cdef MPI_Count scount = 0, rcount = 0
    cdef MPI_Datatype dtype = MPI_BYTE
    cdef object unuseds = pickle_dump(pickle, obj, &sbuf, &scount)
    with nogil: CHKERR( MPI_Sendrecv_c(
            &scount, 1, MPI_COUNT, dst, stag,
            &rcount, 1, MPI_COUNT, src, rtag,
            comm, MPI_STATUS_IGNORE) )
    cdef object unusedr = pickle_alloc(&rbuf, rcount)
    with nogil: CHKERR( MPI_Sendrecv_c(
            sbuf, scount, dtype, dst, stag,
            rbuf, rcount, dtype, src, rtag,
            comm, MPI_STATUS_IGNORE) )
    return pickle_load(pickle, rbuf, rcount)

cdef object PyMPI_bcast_p2p(object obj, int root, MPI_Comm comm):
    cdef Pickle pickle = PyMPI_PICKLE
    cdef void *buf = NULL
    cdef MPI_Count count = 0
    cdef MPI_Datatype dtype = MPI_BYTE
    cdef int rank = MPI_PROC_NULL
    CHKERR( MPI_Comm_rank(comm, &rank) )
    if root == rank: obj = pickle_dump(pickle, obj, &buf, &count)
    with PyMPI_Lock(comm, "@bcast_p2p@"):
        with nogil: CHKERR( MPI_Bcast_c(&count, 1, MPI_COUNT, root, comm) )
        if root != rank: obj = pickle_alloc(&buf, count)
        with nogil: CHKERR( MPI_Bcast_c(buf, count, dtype, root, comm) )
    return pickle_load(pickle, buf, count)

cdef object PyMPI_reduce_p2p(object sendobj, object op, int root,
                             MPI_Comm comm, int tag):
    # Get communicator size and rank
    cdef int size = MPI_UNDEFINED
    cdef int rank = MPI_PROC_NULL
    CHKERR( MPI_Comm_size(comm, &size) )
    CHKERR( MPI_Comm_rank(comm, &rank) )
    # Check root argument
    if root < 0 or root >= size:
        <void>MPI_Comm_call_errhandler(comm, MPI_ERR_ROOT)
        raise MPIException(MPI_ERR_ROOT)
    #
    cdef object result = PyMPI_copy(sendobj)
    cdef object unused
    # Compute reduction at process 0
    cdef unsigned int umask = <unsigned int> 1
    cdef unsigned int usize = <unsigned int> size
    cdef unsigned int urank = <unsigned int> rank
    cdef int target = 0
    while umask < usize:
        if (umask & urank) != 0:
            target = <int> ((urank & ~umask) % usize)
            PyMPI_send_p2p(result, target, tag, comm)
        else:
            target = <int> (urank | umask)
            if target < size:
                unused = PyMPI_recv_p2p(target, tag, comm)
                result = op(result, unused)
        umask <<= 1
    # Send reduction to root
    if root != 0:
        if rank == 0:
            PyMPI_send_p2p(result, root, tag, comm)
        elif rank == root:
            result = PyMPI_recv_p2p(0, tag, comm)
    if rank != root:
        result = None
    #
    return result

cdef object PyMPI_scan_p2p(object sendobj, object op,
                           MPI_Comm comm, int tag):
    # Get communicator size and rank
    cdef int size = MPI_UNDEFINED
    cdef int rank = MPI_PROC_NULL
    CHKERR( MPI_Comm_size(comm, &size) )
    CHKERR( MPI_Comm_rank(comm, &rank) )
    #
    cdef object result  = PyMPI_copy(sendobj)
    cdef object partial = result
    cdef object unused
    # Compute prefix reduction
    cdef unsigned int umask = <unsigned int> 1
    cdef unsigned int usize = <unsigned int> size
    cdef unsigned int urank = <unsigned int> rank
    cdef int target = 0
    while umask < usize:
        target = <int> (urank ^ umask)
        if target < size:
            unused = PyMPI_sendrecv_p2p(
                partial, target, tag, target, tag, comm)
            if rank > target:
                partial = op(unused, partial)
                result = op(unused, result)
            else:
                unused = op(partial, unused)
                partial = unused
        umask <<= 1
    #
    return result

cdef object PyMPI_exscan_p2p(object sendobj, object op,
                             MPI_Comm comm, int tag):
    # Get communicator size and rank
    cdef int size = MPI_UNDEFINED
    cdef int rank = MPI_PROC_NULL
    CHKERR( MPI_Comm_size(comm, &size) )
    CHKERR( MPI_Comm_rank(comm, &rank) )
    #
    cdef object result  = PyMPI_copy(sendobj)
    cdef object partial = result
    cdef object unused
    # Compute prefix reduction
    cdef unsigned int umask = <unsigned int> 1
    cdef unsigned int usize = <unsigned int> size
    cdef unsigned int urank = <unsigned int> rank
    cdef unsigned int uflag = <unsigned int> 0
    cdef int target = 0
    while umask < usize:
        target = <int> (urank ^ umask)
        if target < size:
            unused = PyMPI_sendrecv_p2p(
                partial, target, tag, target, tag, comm)
            if rank > target:
                partial = op(unused, partial)
                if uflag == 0:
                    uflag = 1
                    result = unused
                else:
                    result = op(unused, result)
            else:
                unused = op(partial, unused)
                partial = unused
        umask <<= 1
    #
    if rank == 0:
        result = None
    return result

# ---

cdef extern from * nogil:
    int PyMPI_Commctx_intra(MPI_Comm, MPI_Comm*, int*)
    int PyMPI_Commctx_inter(MPI_Comm, MPI_Comm*, int*, MPI_Comm*, int*)
    int PyMPI_Commctx_finalize()

cdef int PyMPI_Commctx_INTRA(
    MPI_Comm comm,
    MPI_Comm *dupcomm, int *tag,
) except -1:
    with PyMPI_Lock(comm, "@commctx_intra"):
        CHKERR( PyMPI_Commctx_intra(comm, dupcomm, tag) )
    return 0

cdef int PyMPI_Commctx_INTER(
    MPI_Comm comm,
    MPI_Comm *dupcomm, int *tag,
    MPI_Comm *localcomm, int *low_group,
) except -1:
    with PyMPI_Lock(comm, "@commctx_inter"):
        CHKERR( PyMPI_Commctx_inter(comm, dupcomm, tag,
                                    localcomm, low_group) )
    return 0


def _commctx_intra(
    Intracomm comm: Intracomm,
) -> tuple[Intracomm, int]:
    """
    Create/get intracommunicator duplicate.
    """
    cdef int tag = MPI_UNDEFINED
    cdef Intracomm dupcomm = <Intracomm>New(Intracomm)
    PyMPI_Commctx_INTRA(comm.ob_mpi, &dupcomm.ob_mpi, &tag)
    return (dupcomm, tag)


def _commctx_inter(
    Intercomm comm: Intercomm,
) -> tuple[Intercomm, int, Intracomm, bool]:
    """
    Create/get intercommunicator duplicate.
    """
    cdef int tag = MPI_UNDEFINED, low_group = 0
    cdef Intercomm dupcomm = <Intercomm>New(Intercomm)
    cdef Intracomm localcomm = <Intracomm>New(Intracomm)
    PyMPI_Commctx_INTER(comm.ob_mpi, &dupcomm.ob_mpi, &tag,
                        &localcomm.ob_mpi, &low_group)
    return (dupcomm, tag, localcomm, <bint>low_group)


# ---

cdef object PyMPI_reduce_intra(object sendobj, object op,
                               int root, MPI_Comm comm):
    cdef int tag = MPI_UNDEFINED
    PyMPI_Commctx_INTRA(comm, &comm, &tag)
    return PyMPI_reduce_p2p(sendobj, op, root, comm, tag)

cdef object PyMPI_reduce_inter(object sendobj, object op,
                               int root, MPI_Comm comm):
    cdef int tag = MPI_UNDEFINED
    cdef MPI_Comm localcomm = MPI_COMM_NULL
    PyMPI_Commctx_INTER(comm, &comm, &tag, &localcomm, NULL)
    # Get communicator remote size and rank
    cdef int size = MPI_UNDEFINED
    cdef int rank = MPI_PROC_NULL
    CHKERR( MPI_Comm_remote_size(comm, &size) )
    CHKERR( MPI_Comm_rank(comm, &rank) )
    if root >= 0 and root < size:
        # Reduce in local group and send to remote root
        sendobj = PyMPI_reduce_p2p(sendobj, op, 0, localcomm, tag)
        if rank == 0: PyMPI_send_p2p(sendobj, root, tag, comm)
        return None
    elif root == MPI_ROOT:  # Receive from remote group
        return PyMPI_recv_p2p(0, tag, comm)
    elif root == MPI_PROC_NULL:  # This process does nothing
        return None
    else:  # Wrong root argument
        <void>MPI_Comm_call_errhandler(comm, MPI_ERR_ROOT)
        raise MPIException(MPI_ERR_ROOT)


cdef object PyMPI_allreduce_intra(object sendobj, object op, MPI_Comm comm):
    cdef int tag = MPI_UNDEFINED
    PyMPI_Commctx_INTRA(comm, &comm, &tag)
    sendobj = PyMPI_reduce_p2p(sendobj, op, 0, comm, tag)
    return PyMPI_bcast_p2p(sendobj, 0, comm)

cdef object PyMPI_allreduce_inter(object sendobj, object op, MPI_Comm comm):
    cdef int tag = MPI_UNDEFINED
    cdef int rank = MPI_PROC_NULL
    cdef MPI_Comm localcomm = MPI_COMM_NULL
    PyMPI_Commctx_INTER(comm, &comm, &tag, &localcomm, NULL)
    CHKERR( MPI_Comm_rank(comm, &rank) )
    # Reduce in local group, exchange, and broadcast in local group
    sendobj = PyMPI_reduce_p2p(sendobj, op, 0, localcomm, tag)
    if rank == 0:
        sendobj = PyMPI_sendrecv_p2p(sendobj, 0, tag, 0, tag, comm)
    return PyMPI_bcast_p2p(sendobj, 0, localcomm)


cdef object PyMPI_scan_intra(object sendobj, object op, MPI_Comm comm):
    cdef int tag = MPI_UNDEFINED
    PyMPI_Commctx_INTRA(comm, &comm, &tag)
    return PyMPI_scan_p2p(sendobj, op, comm, tag)

cdef object PyMPI_exscan_intra(object sendobj, object op, MPI_Comm comm):
    cdef int tag = MPI_UNDEFINED
    PyMPI_Commctx_INTRA(comm, &comm, &tag)
    return PyMPI_exscan_p2p(sendobj, op, comm, tag)

# ---

cdef inline bint comm_is_intra(MPI_Comm comm) except -1 nogil:
    cdef int inter = 0
    CHKERR( MPI_Comm_test_inter(comm, &inter) )
    if inter: return 0
    else:     return 1


cdef object PyMPI_reduce(object sendobj, object op, int root, MPI_Comm comm):
    if not options.fast_reduce:
        return PyMPI_reduce_naive(sendobj, op, root, comm)
    elif comm_is_intra(comm):
        return PyMPI_reduce_intra(sendobj, op, root, comm)
    else:
        return PyMPI_reduce_inter(sendobj, op, root, comm)


cdef object PyMPI_allreduce(object sendobj, object op, MPI_Comm comm):
    if not options.fast_reduce:
        return PyMPI_allreduce_naive(sendobj, op, comm)
    elif comm_is_intra(comm):
        return PyMPI_allreduce_intra(sendobj, op, comm)
    else:
        return PyMPI_allreduce_inter(sendobj, op, comm)


cdef object PyMPI_scan(object sendobj, object op, MPI_Comm comm):
    if not options.fast_reduce:
        return PyMPI_scan_naive(sendobj, op, comm)
    else:
        return PyMPI_scan_intra(sendobj, op, comm)


cdef object PyMPI_exscan(object sendobj, object op, MPI_Comm comm):
    if not options.fast_reduce:
        return PyMPI_exscan_naive(sendobj, op, comm)
    else:
        return PyMPI_exscan_intra(sendobj, op, comm)

# -----------------------------------------------------------------------------



cdef class Group:
    """
    Group of processes.
    """

    def __cinit__(self, Group group: Group | None = None):
        cinit(self, group)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Group): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_default(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Group:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_Group> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Free` if not null or predefined.
        """
        safefree(self)

    # Group Accessors
    # ---------------

    def Get_size(self) -> int:
        """
        Return the number of processes in a group.
        """
        cdef int size = -1
        CHKERR( MPI_Group_size(self.ob_mpi, &size) )
        return size

    property size:
        """Number of processes."""
        def __get__(self) -> int:
            return self.Get_size()

    def Get_rank(self) -> int:
        """
        Return the rank of this process in a group.
        """
        cdef int rank = MPI_PROC_NULL
        CHKERR( MPI_Group_rank(self.ob_mpi, &rank) )
        return rank

    property rank:
        """Rank of this process."""
        def __get__(self) -> int:
            return self.Get_rank()

    def Translate_ranks(
        self,
        ranks: Sequence[int] | None = None,
        Group group: Group | None = None,
    ) -> list[int]:
        """
        Translate ranks in a group to those in another group.
        """
        cdef MPI_Group group1 = MPI_GROUP_NULL
        cdef MPI_Group group2 = MPI_GROUP_NULL
        cdef int n = 0, *iranks1 = NULL, *iranks2 = NULL
        #
        cdef unused1 = None
        if ranks is not None:
            unused1 = getarray(ranks, &n, &iranks1)
        else:
            CHKERR( MPI_Group_size(self.ob_mpi, &n) )
            unused1 = newarray(n, &iranks1)
            for i in range(n): iranks1[i] = i
        cdef unused2 = newarray(n, &iranks2)
        #
        group1 = self.ob_mpi
        if group is not None:
            group2 = group.ob_mpi
        else:
            CHKERR( MPI_Comm_group(MPI_COMM_WORLD, &group2) )
        try:
            CHKERR( MPI_Group_translate_ranks(
                group1, n, iranks1, group2, iranks2) )
        finally:
            if group is None:
                CHKERR( MPI_Group_free(&group2) )
        #
        cdef object ranks2 = [iranks2[i] for i in range(n)]
        return ranks2

    def Compare(self, Group group: Group) -> int:
        """
        Compare two groups.
        """
        cdef int flag = MPI_UNEQUAL
        CHKERR( MPI_Group_compare(
            self.ob_mpi, group.ob_mpi, &flag) )
        return flag

    # Group Constructors
    # ------------------

    def Dup(self) -> Self:
        """
        Duplicate a group.
        """
        cdef Group group = <Group>New(type(self))
        CHKERR( MPI_Group_union(self.ob_mpi, MPI_GROUP_EMPTY, &group.ob_mpi) )
        return group

    @classmethod
    def Union(cls, Group group1: Group, Group group2: Group) -> Self:
        """
        Create a new group from the union of two existing groups.
        """
        cdef Group group = <Group>New(cls)
        CHKERR( MPI_Group_union(
                group1.ob_mpi, group2.ob_mpi, &group.ob_mpi) )
        return group

    @classmethod
    def Intersection(cls, Group group1: Group, Group group2: Group) -> Self:
        """
        Create a new group from the intersection of two existing groups.
        """
        cdef Group group = <Group>New(cls)
        CHKERR( MPI_Group_intersection(
                group1.ob_mpi, group2.ob_mpi, &group.ob_mpi) )
        return group

    Intersect = Intersection

    @classmethod
    def Difference(cls, Group group1: Group, Group group2: Group) -> Self:
        """
        Create a new group from the difference of two existing groups.
        """
        cdef Group group = <Group>New(cls)
        CHKERR( MPI_Group_difference(
                group1.ob_mpi, group2.ob_mpi, &group.ob_mpi) )
        return group

    def Incl(self, ranks: Sequence[int]) -> Self:
        """
        Create a new group by including listed members.
        """
        cdef int n = 0, *iranks = NULL
        ranks = getarray(ranks, &n, &iranks)
        cdef Group group = <Group>New(type(self))
        CHKERR( MPI_Group_incl(self.ob_mpi, n, iranks, &group.ob_mpi) )
        return group

    def Excl(self, ranks: Sequence[int]) -> Self:
        """
        Create a new group by excluding listed members.
        """
        cdef int n = 0, *iranks = NULL
        ranks = getarray(ranks, &n, &iranks)
        cdef Group group = <Group>New(type(self))
        CHKERR( MPI_Group_excl(self.ob_mpi, n, iranks, &group.ob_mpi) )
        return group

    def Range_incl(self, ranks: Sequence[tuple[int, int, int]]) -> Self:
        """
        Create a new group by including ranges of members.
        """
        cdef int *p = NULL, (*ranges)[3]  # = NULL ## XXX cython fails
        ranges = NULL
        cdef int n = <int>len(ranks)
        cdef unused1 = allocate(n, sizeof(int[3]), &ranges)
        for i in range(n):
            p = <int*> ranges[i]
            p[0], p[1], p[2] = ranks[i]
        cdef Group group = <Group>New(type(self))
        CHKERR( MPI_Group_range_incl(self.ob_mpi, n, ranges, &group.ob_mpi) )
        return group

    def Range_excl(self, ranks: Sequence[tuple[int, int, int]]) -> Self:
        """
        Create a new group by excluding ranges of members.
        """
        cdef int *p = NULL, (*ranges)[3]  # = NULL ## XXX cython fails
        ranges = NULL
        cdef int n = <int>len(ranks)
        cdef unused1 = allocate(n, sizeof(int[3]), &ranges)
        for i in range(n):
            p = <int*> ranges[i]
            p[0], p[1], p[2] = ranks[i]
        cdef Group group = <Group>New(type(self))
        CHKERR( MPI_Group_range_excl(self.ob_mpi, n, ranges, &group.ob_mpi) )
        return group

    @classmethod
    def Create_from_session_pset(
        cls,
        Session session: Session,
        pset_name: str,
    ) -> Self:
        """
        Create a new group from session and process set.
        """
        cdef char *cname = NULL
        pset_name = asmpistr(pset_name, &cname)
        cdef Group group = <Group>New(cls)
        CHKERR( MPI_Group_from_session_pset(
            session.ob_mpi, cname, &group.ob_mpi) )
        return group

    # Group Destructor
    # ----------------

    def Free(self) -> None:
        """
        Free a group.
        """
        cdef MPI_Group save = self.ob_mpi
        CHKERR( MPI_Group_free(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Group_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Group:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_Group_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Group_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Group:
        """
        """
        return fromhandle(MPI_Group_f2c(arg))


cdef Group __GROUP_NULL__  = def_Group ( MPI_GROUP_NULL  , "GROUP_NULL"  )
cdef Group __GROUP_EMPTY__ = def_Group ( MPI_GROUP_EMPTY , "GROUP_EMPTY" )


# Predefined group handles
# ------------------------

GROUP_NULL  = __GROUP_NULL__   #: Null group handle
GROUP_EMPTY = __GROUP_EMPTY__  #: Empty group handle



# http://mvapich.cse.ohio-state.edu/benchmarks/

from mpi4py import MPI


def osu_barrier(
    BENCHMARH="MPI Barrier Latency Test",
    skip=1000,
    loop=10000,
    skip_large=10,
    loop_large=100,
    large_message_size=8192,
    MAX_MSG_SIZE=1 << 20,
):
    comm = MPI.COMM_WORLD
    myid = comm.Get_rank()
    numprocs = comm.Get_size()

    if numprocs < 2:
        if myid == 0:
            errmsg = "This test requires at least two processes"
        else:
            errmsg = None
        raise SystemExit(errmsg)

    if myid == 0:
        print(f"# {BENCHMARH}")
    if myid == 0:
        print(f"# {'Size [B]':<8s}{'Latency [us]':>20s}")

    size = 0
    _ = large_message_size  # unused
    _ = MAX_MSG_SIZE  # unused

    skip = skip_large
    loop = loop_large
    iterations = list(range(loop + skip))
    #
    comm.Barrier()
    for i in iterations:
        if i == skip:
            t_start = MPI.Wtime()
        comm.Barrier()
    t_end = MPI.Wtime()
    comm.Barrier()
    #
    if myid == 0:
        latency = (t_end - t_start) * 1e6 / loop
        print(f"{size:-10d}{latency:20.2f}")


if __name__ == "__main__":
    osu_barrier()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Typing support."""
# pylint: disable=too-few-public-methods

import sys
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    SupportsIndex,
    Tuple,
    TypeVar,
    Union,
)

from .MPI import (
    BottomType,
    Datatype,
    InPlaceType,
)

__all__ = [
    "SupportsBuffer",
    "SupportsDLPack",
    "SupportsCAI",
    "Buffer",
    "Bottom",
    "InPlace",
    "Aint",
    "Count",
    "Displ",
    "Offset",
    "TypeSpec",
    "BufSpec",
    "BufSpecB",
    "BufSpecV",
    "BufSpecW",
    "TargetSpec",
]

_Stream = Union[int, Any]
_PyCapsule = object
_DeviceType = int
_DeviceID = int


class SupportsBuffer(Protocol):
    """Python buffer protocol.

    .. seealso:: :ref:`python:bufferobjects`
    """

    if sys.version_info >= (3, 12):  # pragma: no branch

        def __buffer__(self, flags: int, /) -> memoryview:
            """Create a buffer from a Python object."""


class SupportsDLPack(Protocol):
    """DLPack data interchange protocol.

    .. seealso:: :ref:`dlpack:python-spec`
    """

    def __dlpack__(self, /, *, stream: Optional[_Stream] = None) -> _PyCapsule:
        """Export data for consumption as a DLPack capsule."""

    def __dlpack_device__(self) -> Tuple[_DeviceType, _DeviceID]:
        """Get device type and device ID in DLPack format."""


class SupportsCAI(Protocol):
    """CUDA Array Interface (CAI) protocol.

    .. seealso:: :ref:`numba:cuda-array-interface`
    """

    @property
    def __cuda_array_interface__(self) -> Dict[str, Any]:
        """CAI protocol data."""


Buffer = Union[
    SupportsBuffer,
    SupportsDLPack,
    SupportsCAI,
]
"""
Buffer-like object.
"""


Bottom = Union[BottomType, None]
"""
Start of the address range.
"""


InPlace = Union[InPlaceType, None]
"""
In-place buffer argument.
"""


Aint = SupportsIndex
"""
Address-sized integral type.
"""


Count = SupportsIndex
"""
Integral type for counts.
"""


Displ = SupportsIndex
"""
Integral type for displacements.
"""


Offset = SupportsIndex
"""
Integral type for offsets.
"""


TypeSpec = Union[Datatype, str]
"""
Datatype specification.
"""


BufSpec = Union[
    Buffer,
    Tuple[Buffer, Count],
    Tuple[Buffer, TypeSpec],
    Tuple[Buffer, Count, TypeSpec],
    Tuple[Bottom, Count, Datatype],
    List[Any],
]
"""
Buffer specification.

* `Buffer`
* Tuple[`Buffer`, `Count`]
* Tuple[`Buffer`, `TypeSpec`]
* Tuple[`Buffer`, `Count`, `TypeSpec`]
* Tuple[`Bottom`, `Count`, `Datatype`]
"""


BufSpecB = Union[
    Buffer,
    Tuple[Buffer, Count],
    Tuple[Buffer, TypeSpec],
    Tuple[Buffer, Count, TypeSpec],
    List[Any],
]
"""
Buffer specification (block).

* `Buffer`
* Tuple[`Buffer`, `Count`]
* Tuple[`Buffer`, `TypeSpec`]
* Tuple[`Buffer`, `Count`, `TypeSpec`]
"""


BufSpecV = Union[
    Buffer,
    Tuple[Buffer, Sequence[Count]],
    Tuple[Buffer, Tuple[Sequence[Count], Sequence[Displ]]],
    Tuple[Buffer, TypeSpec],
    Tuple[Buffer, Sequence[Count], TypeSpec],
    Tuple[Buffer, Tuple[Sequence[Count], Sequence[Displ]], TypeSpec],
    Tuple[Buffer, Sequence[Count], Sequence[Displ], TypeSpec],
    Tuple[Bottom, Tuple[Sequence[Count], Sequence[Displ]], Datatype],
    Tuple[Bottom, Sequence[Count], Sequence[Displ], Datatype],
    List[Any],
]
"""
Buffer specification (vector).

* `Buffer`
* Tuple[`Buffer`, Sequence[`Count`]]
* Tuple[`Buffer`, Tuple[Sequence[`Count`], Sequence[`Displ`]]]
* Tuple[`Buffer`, `TypeSpec`]
* Tuple[`Buffer`, Sequence[`Count`], `TypeSpec`]
* Tuple[`Buffer`, Tuple[Sequence[`Count`], Sequence[`Displ`]], `TypeSpec`]
* Tuple[`Buffer`, Sequence[`Count`], Sequence[`Displ`], `TypeSpec`]
* Tuple[`Bottom`, Tuple[Sequence[`Count`], Sequence[`Displ`]], `Datatype`]
* Tuple[`Bottom`, Sequence[`Count`], Sequence[`Displ`], `Datatype`]
"""


BufSpecW = Union[
    Tuple[Buffer, Sequence[Datatype]],
    Tuple[Buffer, Tuple[Sequence[Count], Sequence[Displ]], Sequence[Datatype]],
    Tuple[Buffer, Sequence[Count], Sequence[Displ], Sequence[Datatype]],
    Tuple[Bottom, Tuple[Sequence[Count], Sequence[Displ]], Sequence[Datatype]],
    Tuple[Bottom, Sequence[Count], Sequence[Displ], Sequence[Datatype]],
    List[Any],
]
"""
Buffer specification (generalized).

* Tuple[`Buffer`, Sequence[`Datatype`]]
* Tuple[`Buffer`, \
        Tuple[Sequence[`Count`], Sequence[`Displ`]], Sequence[`Datatype`]]
* Tuple[`Buffer`, Sequence[`Count`], Sequence[`Displ`], Sequence[`Datatype`]]
* Tuple[`Bottom`, \
        Tuple[Sequence[`Count`], Sequence[`Displ`]], Sequence[`Datatype`]]
* Tuple[`Bottom`, Sequence[`Count`], Sequence[`Displ`], Sequence[`Datatype`]]
"""


TargetSpec = Union[
    Displ,
    Tuple[()],
    Tuple[Displ],
    Tuple[Displ, Count],
    Tuple[Displ, Count, TypeSpec],
    List[Any],
]
"""
Target specification.

* `Displ`
* Tuple[()]
* Tuple[`Displ`]
* Tuple[`Displ`, `Count`]
* Tuple[`Displ`, `Count`, `TypeSpec`]
"""


S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")



import pathlib

import cffi

from mpi4py import MPI

_libdir = pathlib.Path(__file__).resolve().parent

ffi = cffi.FFI()
if MPI._sizeof(MPI.Comm) == ffi.sizeof("int"):
    MPI_Comm = "int"
else:
    MPI_Comm = "void*"
ffi.cdef(f"""
typedef {MPI_Comm} MPI_Comm;
void sayhello(MPI_Comm);
""")
lib = ffi.dlopen(_libdir / "libhelloworld.so")


def sayhello(comm):
    comm_c = ffi.cast("MPI_Comm", comm.handle)
    lib.sayhello(comm_c)



import os
import pathlib
import re
import sys


def get_name(settings=None):  # noqa: ARG001
    name = "mpi4py"
    suffix = os.environ.get("MPI4PY_DIST_SUFFIX")
    if suffix:
        name = f"{name}-{suffix}"
    return name


def get_version(settings=None):  # noqa: ARG001
    topdir = pathlib.Path(__file__).resolve().parent.parent
    source = topdir / "src" / "mpi4py" / "__init__.py"
    content = source.read_text(encoding="utf-8")
    m = re.search(r'__version__\s*=\s*"(.*)"', content)
    version = m.groups()[0]
    local_version = os.environ.get("MPI4PY_LOCAL_VERSION")
    if local_version:
        version = f"{version}+{local_version}"
    return version


def get_readme(settings=None):  # noqa: ARG001
    topdir = pathlib.Path(__file__).resolve().parent.parent
    filelist = ("DESCRIPTION.rst", "CITATION.rst", "INSTALL.rst")
    template = "See `{0} <{0}>`_.\n\n"
    template += ".. include:: {0}\n"
    text = template.format(filelist[0])
    for filename in filelist:
        source = topdir / filename
        with source.open(encoding="utf-8") as f:
            includeline = template.format(filename)
            text = text.replace(includeline, f.read())
    return {
        "text": text,
        "content-type": "text/x-rst",
    }


description = "Python bindings for MPI"
requires_python = ">=3.8"
license = "BSD-3-Clause"  # noqa: A001
authors = [
    {"name": "Lisandro Dalcin", "email": "dalcinl@gmail.com"},
]
keywords = [
    "scientific computing",
    "parallel computing",
    "message passing interface",
    "MPI",
]
classifiers = [
    "Development Status :: 6 - Mature",
    "Environment :: GPU",
    "Environment :: GPU :: NVIDIA CUDA",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Operating System :: MacOS",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: POSIX :: BSD",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Unix",
    "Programming Language :: C",
    "Programming Language :: Cython",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Programming Language :: Python :: 3.14",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Scientific/Engineering",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: System :: Distributed Computing",
    "Typing :: Typed",
]
urls = {
    "Homepage":      "https://mpi4py.github.io/mpi4py/",
    "Documentation": "https://mpi4py.readthedocs.io/en/stable/",
    "Source":        "https://github.com/mpi4py/mpi4py",
    "Issues":        "https://github.com/mpi4py/mpi4py/issues",
    "Discussions":   "https://github.com/mpi4py/mpi4py/discussions",
    "Downloads":     "https://github.com/mpi4py/mpi4py/releases",
}  # fmt: skip


def dynamic_metadata(field, settings=None):
    getter = globals().get("get_" + field)
    if getter:
        return getter(settings)
    return globals()[field.replace(".", "_")]


if __name__ == "__main__":
    print(dynamic_metadata(sys.argv[1]))



import importlib
import io
import os
import pathlib
import sys
import unittest
import warnings

import mpi4py

pkgdir = pathlib.Path(mpi4py.__file__).resolve().parent


class TestImport(unittest.TestCase):
    #
    #
    def testImportMPI(self):
        importlib.import_module("mpi4py.MPI")

    def testImportBench(self):
        importlib.import_module("mpi4py.bench")

    def testImportFutures(self):
        importlib.import_module("mpi4py.futures")
        importlib.import_module("mpi4py.futures.server")
        importlib.import_module("mpi4py.futures.__main__")

    def testImportRun(self):
        importlib.import_module("mpi4py.run")
        importlib.import_module("mpi4py.__main__")

    def testImportTyping(self):
        importlib.import_module("mpi4py.typing")

    def testImportUtil(self):
        importlib.import_module("mpi4py.util")
        importlib.import_module("mpi4py.util.dtlib")
        importlib.import_module("mpi4py.util.pkl5")
        importlib.import_module("mpi4py.util.pool")
        importlib.import_module("mpi4py.util.sync")

    def testImportMPIABI(self):
        importlib.import_module("mpi4py._mpiabi")


class TestDataFiles(unittest.TestCase):
    #
    #
    def testTyping(self):
        import importlib.machinery

        py_typed = pkgdir / "py.typed"
        self.assertTrue(py_typed.exists())
        suffixes = [
            *importlib.machinery.SOURCE_SUFFIXES,
            *importlib.machinery.EXTENSION_SUFFIXES,
        ]
        for root, _, files in os.walk(pkgdir):
            root = pathlib.Path(root)
            for fname in files:
                name, _, extra = fname.partition(".")
                suffix = f".{extra}"
                for entry in suffixes:
                    if suffix.endswith(entry):
                        pyi = root / f"{name}.pyi"
                        self.assertTrue(pyi.exists())
                        break

    def testCython(self):
        for pxd in [
            pkgdir / "__init__.pxd",
            pkgdir / "libmpi.pxd",
            pkgdir / "MPI.pxd",
        ]:
            self.assertTrue(pxd.exists())

    def testHeaders(self):
        for hdr in [
            pkgdir / "MPI.h",
            pkgdir / "MPI_api.h",
            pkgdir / "include" / "mpi4py" / "pycapi.h",
            pkgdir / "include" / "mpi4py" / "mpi4py.h",
            pkgdir / "include" / "mpi4py" / "mpi4py.i",
        ]:
            self.assertTrue(hdr.exists())


class TestMPIABI(unittest.TestCase):
    #
    #
    def setUp(self):
        module = importlib.import_module("mpi4py.MPI")
        self.mpiabi = importlib.import_module("mpi4py._mpiabi")
        self.sentinel = getattr(self.mpiabi._get_mpiabi, "mpiabi", None)
        self.savedstate = (self.mpiabi.MPIABI, self.mpiabi.LIBMPI)
        if os.name == "posix":
            self.libmpi = module.__file__
        else:
            vendor = module.get_vendor()[0]
            if vendor == "Intel MPI":
                self.libmpi = "impi.dll"
            elif vendor == "Microsoft MPI":
                self.libmpi = "msmpi.dll"

    def tearDown(self):
        self.mpiabi._registry.pop("mpi4py.xyz", None)
        if self.sentinel is not None:
            self.mpiabi._get_mpiabi.mpiabi = self.sentinel
        elif hasattr(self.mpiabi._get_mpiabi, "mpiabi"):
            del self.mpiabi._get_mpiabi.mpiabi
        self.mpiabi.MPIABI, self.mpiabi.LIBMPI = self.savedstate
        del self.mpiabi

    def testGetFromString(self):
        mpiabi = self.mpiabi
        mpiabi._get_mpiabi.mpiabi = None
        mpiabi.MPIABI = "mpiabi"
        result = mpiabi._get_mpiabi()
        self.assertEqual(result, "mpiabi")

    def testGetFromLibMPI(self):
        mpiabi = self.mpiabi
        mpiabi._get_mpiabi.mpiabi = None
        mpiabi.LIBMPI = self.libmpi
        result = mpiabi._get_mpiabi()
        expected = {"mpiabi", "mpich", "openmpi", "impi", "msmpi"}
        self.assertIn(result, expected)

    def testString(self):
        mpiabi = self.mpiabi
        posix = os.name == "posix"
        for string, expected in (
            ("mpiabi", "mpiabi"),
            ("MPICH", "mpich" if posix else "impi"),
            ("I_MPI", "mpich" if posix else "impi"),
            ("Open MPI", "openmpi"),
            ("OPEN-MPI", "openmpi"),
            ("", ""),
        ):
            result = mpiabi._get_mpiabi_from_string(string)
            self.assertEqual(result, expected)

    def testLibMPI(self):
        mpiabi = self.mpiabi
        libmpi = self.libmpi
        result = mpiabi._get_mpiabi_from_libmpi(libmpi)
        expected = {"mpiabi", "mpich", "openmpi", "impi", "msmpi"}
        self.assertIn(result, expected)

    def testSuffix(self):
        mpiabi = self.mpiabi
        for abiname in ("mpiabi1", "mpiabi2", ""):
            mpiabi._get_mpiabi.mpiabi = abiname
            result = mpiabi._get_mpiabi_suffix("mpi4py.xyz")
            self.assertIsNone(result)
            mpiabi._register("mpi4py.xyz", abiname)
            mpiabi._register("mpi4py.xyz", abiname)
            result = mpiabi._get_mpiabi_suffix("mpi4py.xyz")
            self.assertEqual(result, f".{abiname}" if abiname else "")

    def testFinder(self):
        mpiabi = self.mpiabi
        finder = mpiabi._Finder
        registry = mpiabi._registry
        if finder not in sys.meta_path:
            mpiabi._install_finder()
            mpiabi._install_finder()
            self.assertIs(finder, sys.meta_path.pop())
        if "mpi4py.MPI" not in registry:
            mpiabi._get_mpiabi.mpiabi = ""
            mpiabi._register("mpi4py.MPI", "")
        spec = finder.find_spec("mpi4py.MPI", mpi4py.__path__)
        if mpiabi._get_mpiabi.mpiabi == "":
            del mpiabi._get_mpiabi.mpiabi
            del registry["mpi4py.MPI"]
        self.assertIsNotNone(spec)
        #
        spec = finder.find_spec("mpi4py.xyz", mpi4py.__path__)
        self.assertIsNone(spec)
        mpiabi._get_mpiabi.mpiabi = ""
        mpiabi._register("mpi4py.xyz", "")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            spec = finder.find_spec("mpi4py.xyz", mpi4py.__path__)
            self.assertIsNone(spec)
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            with self.assertRaises(RuntimeWarning):
                finder.find_spec("mpi4py.xyz", mpi4py.__path__)

    def testVerboseInfo(self):
        mpiabi = self.mpiabi
        message = "@message@"
        output = f"# [{mpiabi.__name__}] {message}\n"
        sys.stderr = stderr = io.StringIO()
        try:
            mpiabi._verbose_info(message, verbosity=-1)
        finally:
            sys.stderr = sys.__stderr__
        self.assertEqual(stderr.getvalue(), output)

    def testDLOpen(self):
        mpiabi = self.mpiabi
        os_name_save = os.name
        os_environ_save = os.environ
        sys_platform_save = sys.platform
        try:
            for os_name, sys_platform in (
                ("posix", "linux"),
                ("posix", "freebsd"),
                ("posix", "darwin"),
                ("nt", "win32"),
            ):
                os.name = os_name
                sys.platform = sys_platform
                mpiabi._dlopen_rpath()
            os.environ = env = {}
            env["I_MPI_ROOT"] = "/usr"
            mpiabi._dlopen_rpath()
            env["MSMPI_ROOT"] = "/usr"
            mpiabi._dlopen_rpath()
            env["MSMPI_BIN"] = "/usr/bin"
            mpiabi._dlopen_rpath()
        finally:
            os.name = os_name_save
            os.environ = os_environ_save  # noqa: B003
            sys.platform = sys_platform_save


if __name__ == "__main__":
    unittest.main()



import sys
from typing import Any, Iterable, Sequence, overload

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from .. import MPI
from ..MPI import (
    ANY_SOURCE as ANY_SOURCE,
    ANY_TAG as ANY_TAG,
    PROC_NULL as PROC_NULL,
    ROOT as ROOT,
    Pickle as Pickle,
    Status as Status,
)
from ..typing import Buffer

pickle: Pickle = ...

class _BigMPI:
    blocksize: int = ...
    cache: dict[int, MPI.Datatype] = ...
    def __init__(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *exc: object) -> None: ...
    def __call__(self, buf: Buffer) -> tuple[Buffer, int, MPI.Datatype]: ...

_bigmpi: _BigMPI = ...

class Request(tuple[MPI.Request, ...]):
    @overload
    def __new__(cls, request: MPI.Request | None = None) -> Request: ...
    @overload
    def __new__(cls, request: Iterable[MPI.Request]) -> Request: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __bool__(self) -> bool: ...
    def Free(self) -> None: ...
    def free(self) -> None: ...
    def cancel(self) -> None: ...
    def get_status(
        self,
        status: Status | None = None,
    ) -> bool: ...
    def test(
        self,
        status: Status | None = None,
    ) -> tuple[bool, Any | None]: ...
    def wait(
        self,
        status: Status | None = None,
    ) -> Any: ...
    @classmethod
    def get_status_all(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> bool: ...
    @classmethod
    def testall(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> tuple[bool, list[Any] | None]: ...
    @classmethod
    def waitall(
        cls,
        requests: Sequence[Request],
        statuses: list[Status] | None = None,
    ) -> list[Any]: ...

class Message(tuple[MPI.Message, ...]):
    @overload
    def __new__(cls, message: MPI.Message | None = None) -> Message: ...
    @overload
    def __new__(cls, message: Iterable[MPI.Message]) -> Message: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __bool__(self) -> bool: ...
    def free(self) -> None: ...
    def recv(self, status: Status | None = None) -> Any: ...
    def irecv(self) -> Request: ...
    @classmethod
    def probe(
        cls,
        comm: MPI.Comm,
        source: int = ANY_SOURCE,
        tag: int = ANY_TAG,
        status: Status | None = None,
    ) -> Message: ...
    @classmethod
    def iprobe(
        cls,
        comm: MPI.Comm,
        source: int = ANY_SOURCE,
        tag: int = ANY_TAG,
        status: Status | None = None,
    ) -> Message | None: ...

class Comm(MPI.Comm):
    def __new__(cls, comm: MPI.Comm | None = None) -> Self: ...
    def send(self, obj: Any, dest: int, tag: int = 0) -> None: ...
    def bsend(self, obj: Any, dest: int, tag: int = 0) -> None: ...
    def ssend(self, obj: Any, dest: int, tag: int = 0) -> None: ...
    def isend(self, obj: Any, dest: int, tag: int = 0) -> Request: ...  # type: ignore[override]
    def ibsend(self, obj: Any, dest: int, tag: int = 0) -> Request: ...  # type: ignore[override]
    def issend(self, obj: Any, dest: int, tag: int = 0) -> Request: ...  # type: ignore[override]
    def recv(
        self,
        buf: Buffer | None = None,
        source: int = ANY_SOURCE,
        tag: int = ANY_TAG,
        status: Status | None = None,
    ) -> Any: ...
    def irecv(  # type: ignore[override]
        self,
        buf: Buffer | None = None,
        source: int = ANY_SOURCE,
        tag: int = ANY_TAG,
    ) -> Request: ...
    def sendrecv(
        self,
        sendobj: Any,
        dest: int,
        sendtag: int = 0,
        recvbuf: Buffer | None = None,
        source: int = ANY_SOURCE,
        recvtag: int = ANY_TAG,
        status: Status | None = None,
    ) -> Any: ...
    def mprobe(  # type: ignore[override]
        self,
        source: int = ANY_SOURCE,
        tag: int = ANY_TAG,
        status: Status | None = None,
    ) -> Message: ...
    def improbe(  # type: ignore[override]
        self,
        source: int = ANY_SOURCE,
        tag: int = ANY_TAG,
        status: Status | None = None,
    ) -> Message | None: ...
    def bcast(
        self,
        obj: Any,
        root: int = 0,
    ) -> Any: ...
    def gather(
        self,
        sendobj: Any,
        root: int = 0,
    ) -> list[Any] | None: ...
    def scatter(
        self,
        sendobj: Sequence[Any] | None,
        root: int = 0,
    ) -> Any: ...
    def allgather(
        self,
        sendobj: Any,
    ) -> list[Any]: ...
    def alltoall(
        self,
        sendobj: Sequence[Any],
    ) -> list[Any]: ...

class Intracomm(Comm, MPI.Intracomm): ...
class Intercomm(Comm, MPI.Intercomm): ...



import sys

import mpiunittest as unittest

from mpi4py import MPI


class TestInfoNull(unittest.TestCase):
    #
    def testTruth(self):
        self.assertFalse(bool(MPI.INFO_NULL))

    def testPickle(self):
        from pickle import dumps, loads

        null = loads(dumps(MPI.INFO_NULL))
        self.assertIs(null, MPI.INFO_NULL)
        null = loads(dumps(MPI.Info()))
        self.assertIsNot(null, MPI.INFO_NULL)
        self.assertEqual(null, MPI.INFO_NULL)

    def testPyMethods(self):
        inull = MPI.INFO_NULL

        def getitem():
            return inull["k"]

        def setitem():
            inull["k"] = "v"

        def delitem():
            del inull["k"]

        def update():
            inull.update([])

        def pop():
            inull.pop("k")

        def popitem():
            inull.popitem()

        self.assertEqual(len(inull), 0)
        self.assertNotIn("key", inull)
        self.assertRaises(KeyError, getitem)
        self.assertRaises(KeyError, setitem)
        self.assertRaises(KeyError, delitem)
        self.assertRaises(KeyError, update)
        self.assertRaises(KeyError, pop)
        self.assertRaises(KeyError, popitem)
        self.assertIsNone(inull.get("key", None))
        self.assertIsNone(inull.pop("key", None))
        self.assertEqual(inull.keys(), [])
        self.assertEqual(inull.values(), [])
        self.assertEqual(inull.items(), [])
        self.assertEqual(inull.copy(), inull)
        inull.clear()


class TestInfoEnv(unittest.TestCase):
    #
    KEYS = (
        "command",
        "argv",
        "maxprocs",
        "soft",
        "host",
        "arch",
        "wdir",
        "file",
        "thread_level",
    )

    def testTruth(self):
        self.assertTrue(bool(MPI.INFO_ENV))

    def testPickle(self):
        from pickle import dumps, loads

        ienv = loads(dumps(MPI.INFO_ENV))
        self.assertIs(ienv, MPI.INFO_ENV)
        ienv = loads(dumps(MPI.Info(MPI.INFO_ENV)))
        self.assertIsNot(ienv, MPI.INFO_ENV)
        self.assertEqual(ienv, MPI.INFO_ENV)

    def testPyMethods(self):
        env = MPI.INFO_ENV
        for key in self.KEYS:
            v = env.Get(key)
            del v

    def testDup(self):
        env = MPI.INFO_ENV
        dup = env.Dup()
        try:
            for key in self.KEYS:
                self.assertEqual(env.Get(key), dup.Get(key))
        finally:
            dup.Free()

    def testCreateEnv(self):
        try:
            env = MPI.Info.Create_env()
        except NotImplementedError:
            if MPI.Get_version() >= (4, 0):
                raise
            raise unittest.SkipTest("mpi-info-create-env") from None
        for key in self.KEYS:
            v = env.Get(key)
            del v
        try:
            dup = env.Dup()
            try:
                for key in self.KEYS:
                    self.assertEqual(env.Get(key), dup.Get(key))
            finally:
                dup.Free()
        finally:
            env.Free()
        for args in (
            None,
            [],
            (),
            sys.executable,
            [sys.executable],
            (sys.executable,),
        ):
            MPI.Info.Create_env(args).Free()
            MPI.Info.Create_env(args=args).Free()


class TestInfo(unittest.TestCase):
    #
    def setUp(self):
        self.INFO = MPI.Info.Create()

    def tearDown(self):
        self.INFO.Free()
        self.assertEqual(self.INFO, MPI.INFO_NULL)
        self.INFO = None

    def testTruth(self):
        self.assertTrue(bool(self.INFO))

    def testCreate(self):
        data = {"key1": "value1", "key2": "value2"}
        for items in (None, {}, [], data, list(data.items())):
            info = MPI.Info.Create(items)
            if items is not None:
                self.assertEqual(info.Get_nkeys(), len(items))
                for k, v in dict(items).items():
                    self.assertEqual(info.Get(k), v)
            info.Free()

    def testCreateBad(self):
        with self.assertRaises(TypeError):
            MPI.Info.Create(items=123)

    def testDup(self):
        info = self.INFO.Dup()
        self.assertNotEqual(self.INFO, info)
        self.assertEqual(info.Get_nkeys(), 0)
        info.Free()
        self.assertFalse(info)

    def testGet(self):
        value = self.INFO.Get("key")
        self.assertIsNone(value)

    def testGetNKeys(self):
        self.assertEqual(self.INFO.Get_nkeys(), 0)

    def testGetSetDelete(self):
        INFO = self.INFO
        self.assertEqual(INFO.Get_nkeys(), 0)
        INFO.Set("key", "value")
        nkeys = INFO.Get_nkeys()
        self.assertEqual(nkeys, 1)
        key = INFO.Get_nthkey(0)
        self.assertEqual(key, "key")
        value = INFO.Get("key")
        self.assertEqual(value, "value")
        INFO.Delete("key")
        nkeys = INFO.Get_nkeys()
        self.assertEqual(nkeys, 0)
        value = INFO.Get("key")
        self.assertIsNone(value)

    def testPickle(self):
        from pickle import dumps, loads

        mold = self.INFO
        info = loads(dumps(mold))
        self.assertIsNot(info, mold)
        self.assertNotEqual(info, mold)
        self.assertEqual(info.items(), mold.items())
        info.Free()
        mold = self.INFO
        mold.update([("foo", "bar"), ("answer", "42")])
        info = loads(dumps(mold))
        self.assertIsNot(info, mold)
        self.assertNotEqual(info, mold)
        self.assertEqual(info.items(), mold.items())
        info.Free()

    def testPyMethods(self):
        INFO = self.INFO

        self.assertEqual(len(INFO), 0)
        self.assertNotIn("key", INFO)
        self.assertEqual(INFO.keys(), [])
        self.assertEqual(INFO.values(), [])
        self.assertEqual(INFO.items(), [])

        INFO["key"] = "value"
        self.assertEqual(len(INFO), 1)
        self.assertIn("key", INFO)
        self.assertEqual(INFO["key"], "value")
        for key in INFO:
            self.assertEqual(key, "key")
        self.assertEqual(INFO.keys(), ["key"])
        self.assertEqual(INFO.values(), ["value"])
        self.assertEqual(INFO.items(), [("key", "value")])
        self.assertEqual(key, "key")
        del INFO["key"]
        self.assertEqual(len(INFO), 0)

        INFO["key"] = "value"
        self.assertEqual(INFO.pop("key"), "value")
        self.assertEqual(len(INFO), 0)
        self.assertEqual(INFO.pop("key", "value"), "value")
        self.assertRaises(KeyError, INFO.pop, "key")
        INFO["key1"] = "value1"
        INFO["key2"] = "value2"
        self.assertEqual(INFO.pop("key1"), "value1")
        self.assertEqual(len(INFO), 1)
        self.assertEqual(INFO.pop("key2"), "value2")
        self.assertEqual(len(INFO), 0)

        INFO["key"] = "value"
        self.assertEqual(INFO.popitem(), ("key", "value"))
        self.assertEqual(len(INFO), 0)
        self.assertRaises(KeyError, INFO.popitem)
        INFO["key1"] = "value1"
        INFO["key2"] = "value2"
        self.assertEqual(INFO.popitem(), ("key2", "value2"))
        self.assertEqual(len(INFO), 1)
        self.assertEqual(INFO.popitem(), ("key1", "value1"))
        self.assertEqual(len(INFO), 0)

        self.assertEqual(len(INFO), 0)
        self.assertNotIn("key", INFO)
        self.assertEqual(INFO.keys(), [])
        self.assertEqual(INFO.values(), [])
        self.assertEqual(INFO.items(), [])

        def getitem():
            INFO["key"]

        self.assertRaises(KeyError, getitem)

        def delitem():
            del INFO["key"]

        self.assertRaises(KeyError, delitem)

        INFO.clear()
        INFO.update([("key1", "value1")])
        self.assertEqual(len(INFO), 1)
        self.assertEqual(INFO["key1"], "value1")
        self.assertEqual(INFO.get("key1"), "value1")
        self.assertIsNone(INFO.get("key2"))
        self.assertEqual(INFO.get("key2", "value2"), "value2")
        INFO.update(key2="value2")
        self.assertEqual(len(INFO), 2)
        self.assertEqual(INFO["key1"], "value1")
        self.assertEqual(INFO["key2"], "value2")
        self.assertEqual(INFO.get("key1"), "value1")
        self.assertEqual(INFO.get("key2"), "value2")
        self.assertIsNone(INFO.get("key3"))
        self.assertEqual(INFO.get("key3", "value3"), "value3")
        INFO.update([("key1", "newval1")], key2="newval2")
        self.assertEqual(len(INFO), 2)
        self.assertEqual(INFO["key1"], "newval1")
        self.assertEqual(INFO["key2"], "newval2")
        self.assertEqual(INFO.get("key1"), "newval1")
        self.assertEqual(INFO.get("key2"), "newval2")
        self.assertIsNone(INFO.get("key3"))
        self.assertEqual(INFO.get("key3", "newval3"), "newval3")
        INFO.update({"key1": "val1", "key2": "val2", "key3": "val3"})
        self.assertEqual(len(INFO), 3)
        self.assertEqual(INFO["key1"], "val1")
        self.assertEqual(INFO["key2"], "val2")
        self.assertEqual(INFO["key3"], "val3")
        dupe = INFO.copy()
        self.assertEqual(INFO.items(), dupe.items())
        dupe.Free()
        INFO.clear()
        self.assertEqual(len(INFO), 0)
        self.assertIsNone(INFO.get("key1"))
        self.assertIsNone(INFO.get("key2"))
        self.assertIsNone(INFO.get("key3"))
        self.assertEqual(INFO.get("key1", "value1"), "value1")
        self.assertEqual(INFO.get("key2", "value2"), "value2")
        self.assertEqual(INFO.get("key3", "value3"), "value3")


try:
    MPI.Info.Create().Free()
except NotImplementedError:
    unittest.disable(TestInfo, "mpi-info")
    unittest.disable(TestInfoNull, "mpi-info")
if MPI.Get_version() < (3, 0) or MPI.INFO_ENV == MPI.INFO_NULL:
    unittest.disable(TestInfoEnv, "mpi-info-env")


if __name__ == "__main__":
    unittest.main()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com

# pylint: disable=unused-import
# pylint: disable=redefined-builtin
# pylint: disable=missing-module-docstring

from concurrent.futures import (  # noqa: F401
    ALL_COMPLETED,
    FIRST_COMPLETED,
    FIRST_EXCEPTION,
    BrokenExecutor,
    CancelledError,
    Executor,
    Future,
    InvalidStateError,
    TimeoutError,  # noqa: A004
    as_completed,
    wait,
)
